<?php
/**
 * This file is part of the JustKnit package.
 *
 * @copyright JustKnit (Iris Hammer) <iris@justknit.de>
 * @license All rights reserved (@https://dejure.org/gesetze/UrhG/69a.html)
 * Created on 03-20-2023
 *
 * @package justknit
 */

namespace JustKnit\Controllers\Calculate;

use MODX\Revolution\modX;
use MODX\Revolution\modUser;
use MODX\Revolution\modUserProfile;
use MODX\Revolution\modUserSetting;
use JustKnit\Model\JustKnitClientMeta;
use JustKnit\Model\JustKnitCalc;
use JustKnit\Model\JustKnitDim;
use JustKnit\Controllers\Calculate\Base;

class Calculate extends Base {
    /** @var array $hiddenFlds */
    public $hiddenFlds = [];
   /** @var string $hiddenPls */
    public $hiddenPls = '';
    /** @var array $fields An array of key->name pairs storing the fields passed */
    public $fields = [];

    public function initialize() {
        $this->setDefaultProperties(array(
            'loginContext' => $this->modx->context->get('key'),
            'submitVar' => 'addCalc',
            'tplShow' => 'jkInputShow',
            'tplHidden' => 'jkInputHidden',
            'fldName' => 'fn',
            'basisGroup' => 'jkBasis',
            'fieldPrefix' => 'jk',
            'valuePrefix' => 'jkv',
        ));
        $this->modx->lexicon->load('justknit:calc');
        $this->isAuthenticated = $this->modx->user->isAuthenticated($this->getProperty('loginContext'));
    }
    /**
     * Process the controller
     * @return string
     */
    public function process() {
        // prevent output of FormField if User is not logged in
        if (!$this->isAuthenticated) {
            return '';
        }
        $this->showFlds = explode(',', $this->getProperty('show',[]));
        $this->setRequestedFlds('Calcs', array_keys($this->modx->getFields(JustKnitCalc::class)));
        $this->setRequestedFlds('Dims',array_keys($this->modx->getFields(JustKnitDim::class)));
        foreach($this->hiddenFlds as $k => $v) {
            $this->hiddenPls .= '[[+'.$k.']]';
        }
        // set all fields not shown in Form to one placeholder
        $this->modx->toPlaceholder('hiddenflds',$this->hiddenPls, $this->getProperty('fieldPrefix'));
        $userId = $this->modx->user->get('id');
        $this->user = $this->getUserById($userId);
        $this->mainGroup = $this->getMainGroup();
        // Do not load Calc for UserGroup jkBasis (they only caclulate in Javascript)
        if (!$this->mainGroup || $this->mainGroup == $this->getProperty('basisGroup')) {
            return '';
        }
        $this->calc = $this->getCalc();
        if ($this->calc) {
            // if there is an existing calc for this user/resource load calcs and dims
            $this->fields = $this->getCalcFields();
            $this->modx->toPlaceholders($this->fields, $this->getProperty('valuePrefix'));
        } else {
            // for new calc set default values
            $this->setDefaultCalcFields();
        }
        // check for $_POST
        if (!$this->hasPost()) {
            return '';
        }
        $this->fields = $this->dictionary->toArray();
        $this->fields = array_merge($this->getProperties(),$this->fields);
        // existing calc in Form
        if ($this->fields['cid']) {
            $result = $this->runProcessor('UpdateCalc');
        } else {
            $result = $this->runProcessor('CreateCalc');
        }
        $url = $this->modx->makeUrl($this->modx->resource->get('id'));
        $this->modx->sendRedirect($url);
        return '';
    }
    /**
     * Set the fields of Calculation from custom tables.
     *
     * @access public
     * @return void
     */
    public function setRequestedFlds(string $reqFlds, array $tblFlds) {
        foreach(explode(',', $this->getProperty($reqFlds, [])) as $k => $v) {
            $v = trim($v);
            foreach($tblFlds as $tfd) {
                $fn = $v.'_'.$tfd;
                $ts = $this->getChunk($this->getProperty('tplShow'),array($this->getProperty('fldName') => $fn));
                $th = $this->getChunk($this->getProperty('tplHidden'),array($this->getProperty('fldName') => $fn));
                // set all tablefields to placeholders depend on shown or not
                if (in_array($fn,$this->showFlds)){
                    $this->modx->toPlaceholder($fn,$ts, $this->getProperty('fieldPrefix'));
                } else {
                    $this->modx->toPlaceholder($fn,$th, $this->getProperty('fieldPrefix'));
                    $this->hiddenFlds[$this->getProperty('fieldPrefix').'.'.$fn] = $th;
                }
            }
            $res = $reqFlds == 'Calcs' ? $this->modx->setPlaceholder('resultClass','') : '' ;
        }
    }
    /**
     * Set the value fields of calculation  and dimensions to default Values.
     *
     * @access public
     * @return void
     */
    public function setDefaultCalcFields() {
        $userId = $this->user->get('id');
        $df = [];
        $this->defFields = [
            'resource_id'      => $this->modx->resource->get('id'),
            'ccreatedby'        => $userId,
            'calcedby'          => $userId,
            'editedby'         => $userId,
            'createdon'         => date('Y-m-d H:i:s'),
            'editedon'          => date('Y-m-d H:i:s'),
            'dtn_d'             => '',
        ];
        // for each Calc in ScriptProperty &Calcs
        foreach(explode(',', $this->getProperty('Calcs', [])) as $ck => $cv) {
            $cv = trim($cv);
            // set hidden field [calc_calcname] to requested value
            $this->modx->toPlaceholder($cv.'_calcname',$cv,$this->getProperty('valuePrefix'));
            foreach ($this->defFields as $k => $v) {
                $df[$cv.'_'.$k] = $v;
            }
        }
        // for each Dim in ScriptProperty &Dims
        foreach(explode(',', $this->getProperty('Dims', [])) as $dk => $dv) {
            $dv = trim($dv);
            // set hidden field [dim_dimname] to requested value
            $this->modx->toPlaceholder($dv.'_dimname',$dv,$this->getProperty('valuePrefix'));
            foreach ($this->defFields as $k => $v) {
                $df[$dv.'_'.$k] = $v;
                if ($k == 'dtn_d') {
                    if (str_starts_with($dv, 'w') || str_starts_with($dv, 'h')) {
//                        $dtn = str_split($dv)[0];
//                        $this->modx->toPlaceholder($dv.'_'.$k,$dtn,$this->getProperty('valuePrefix'));
//                        $this->defFields[$k] = $dtn;
                        // add direction of dimension depend on first letter
                        $df[$dv.'_'.$k] = str_split($dv)[0];
                    }
                }
            }
        }
        foreach ($df as $f => $g) {
            $sf = $this->modx->getPlaceholder($f);
            if ( !isset($sf) || $sf == 0 || $sf == '' || $sf == NULL) {
                $this->modx->toPlaceholder($f,$g,$this->getProperty('valuePrefix'));
            }
        }
    }
    /**
     * Get the fields of existing Calculation.
     *
     * @access public
     * @return void
     */
    public function getCalcFields() {
        $calcFields = [];
        $dimFields = [];
        $fields = [];
        if (!is_object($this->calc)) {
            $this->modx->log(modX::LOG_LEVEL_ERROR,'$this->calc ist nicht vorhanden');
            return '';
        }
        if (in_array($this->calc->get('calcname'), explode(',', $this->getProperty('Calcs', [])))) {
            $calcFields[$this->calc->get('calcname')] = $this->calc->toArray();
            $this->getDims($this->calc->get('id'));
            if ($this->dims) {
            foreach ($this->dims as $dim) {
                if (!is_object($dim)) {
                    $this->modx->log(modX::LOG_LEVEL_ERROR,'$dim ist nicht vorhanden');
                    return '';
                }
                if (in_array($dim->get('dimname'), explode(',', $this->getProperty('Dims', [])))) {
                    $dimFields[$dim->get('dimname')] = $dim->toArray();
                }
            }
            } else {
                $dimFields=[];
            }
        }
        $f = array_merge($calcFields, $dimFields);
        foreach ($f as $key => $val) {
            // for every calc and every dim one key => array() of fields
            foreach ($val as $k => $v ) {
                // for each field of calc/dim there are fieldnames to generate and field values to attach
                if (is_array($v)) {
                    $fields[$key.'_'.$k] = $this->modx->toJson($v);
                } else {
                    $fields[$key.'_'.$k] = $v;
                }
            }
        }
        // merge values from pattern_meta to fields
        if ($this->calc->get('pattern_meta')) {
            $fields = array_merge($fields,array_diff_key($this->calc->get('pattern_meta'),$fields));
        }
        // set placeholder to identify all calcs and dims belongs to this id
        $this->modx->toPlaceholder('cid',$this->calc->get('id'), $this->getProperty('valuePrefix'));
//$this->modx->log( modX::LOG_LEVEL_ERROR,'[JustKnit] getCalcFields() $fields: '.print_r($fields,1));        
        return $fields;
    }
}
