<?php
/**
 * This file is part of the JustKnit package.
 *
 * @copyright JustKnit (Iris Hammer) <iris@justknit.de>
 * @license All rights reserved (@https://dejure.org/gesetze/UrhG/69a.html)
 * Created on 03-20-2023
 *
 * @package justknit
 */

namespace JustKnit\Controllers\Register;

use MODX\Revolution\modX;
use MODX\Revolution\modUser;
use MODX\Revolution\modUserProfile;
use MODX\Revolution\modUserSetting;
use JustKnit\Model\JustKnitClientMeta;
use JustKnit\Controllers\Register\Base;

class ClientChangePassword extends Base {
    /** @var modUserProfile $profile */
    public $profile;
    /** @var array $errors */
    public $errors = array();

    public function initialize() {
        $this->modx->lexicon->load('justknit:register');
        $this->setDefaultProperties(array(
            'fieldConfirmNewPassword' => 'password_new_confirm',
            'fieldNewPassword' => 'password_new',
            'fieldOldPassword' => 'password_old',
            'placeholderPrefix' => '',
            'preHooks' => '',
            'redirectToLogin' => true,
            'reloadOnSuccess' => false,
            'reloadOnSuccessVar' => 'justknit-success',
            'submitVar' => 'change-password',
            'successMessage' => $this->modx->lexicon('justknit.password_changed_success'),
            'validate' => 'nospam:blank',
            'validateOldPassword' => true,
            'errTpl' => 'lgnErrTpl',
        ));
    }  

    /**
     * Process the controller
     * @return string
     */
    public function process() {
        if (!$this->verifyAuthentication()) return '';

        $this->getProfile();
        if (empty($this->profile)) return '';

        /* if a submission has occurred */
        if (!empty($_POST) && isset($_POST[$this->getProperty('submitVar','change-password')])) {
//$this->modx->log( modX::LOG_LEVEL_ERROR,'[JustKnit] Reset Und??: process() $placeholders '. print_r($_POST,1));
            
            $this->handlePost();
        }
        return '';
    }

    /**
     * Get the Profile of the active User
     * @return modUserProfile
     */
    public function getProfile() {
        $this->profile = $this->modx->user->getOne('Profile');
        if (empty($this->profile)) {
            $this->modx->log(modX::LOG_LEVEL_ERROR,'Could not find profile for user: '.$this->modx->user->get('username'));
        } else {
            $placeholders = array_merge($this->profile->toArray(),array(
                'username' => $this->modx->user->get('username'),
                'id' => $this->modx->user->get('id'),
            ));
            $placeholderPrefix = rtrim($this->getProperty('placeholderPrefix', 'justknit.'), '.');
            $this->modx->toPlaceholders($placeholders, $placeholderPrefix);
//$this->modx->log( modX::LOG_LEVEL_ERROR,'[JustKnit] Reset Und??: process() $placeholders '. print_r($placeholders,1));
            
        }
        return $this->profile;
    }

    /**
     * See if the user has permission to change their password. If not, redirect.
     * @return boolean
     */
    public function verifyAuthentication() {
        $verified = true;
        /* verify authenticated status */
        if (!$this->modx->user->hasSessionContext($this->modx->context->get('key'))) {
            if ($this->getProperty('redirectToLogin',true,'isset')) {
                $this->modx->sendUnauthorizedPage();
            }
            $verified = false;
        }
        return $verified;
    }

    /**
     * Remove the submitVar from the fields array
     * @return void
     */
    public function removeSubmitVar() {
        $submitVar = $this->getProperty('submitVar','change-password');
//this->modx->log( modX::LOG_LEVEL_ERROR,'[JustKnit] Reset Und??: $submitVar '. $submitVar);

        if (!empty($submitVar)) {
            $this->dictionary->remove($submitVar);
        }
    }

    /**
     * Validate the form with FormIt-style validation
     * @return array
     */
    public function validate() {
        $this->validator = $this->register->loadValidator();
        $fields = $this->validator->validateFields($this->dictionary,$this->getProperty('validate',''));
//$this->modx->log( modX::LOG_LEVEL_ERROR,'[JustKnit] Reset Und??: $fields '. print_r($fields,1));

        $passwordFields = array(
            'fieldConfirmNewPassword' => $this->getProperty('fieldConfirmNewPassword','password_new_confirm'),
            'fieldNewPassword' => $this->getProperty('fieldNewPassword','password_new'),
            'fieldOldPassword' => $this->getProperty('fieldOldPassword','password_old'),
        );
//$this->modx->log( modX::LOG_LEVEL_ERROR,'[JustKnit] Reset Und??: $passwordFields '. print_r($passwordFields,1));

        foreach ($fields as $k => $v) {
            if(!in_array($k,$passwordFields)) {
                $fields[$k] = str_replace(array('[', ']'), array('&#91;', '&#93;'), $v);
            }
        }
//$this->modx->log( modX::LOG_LEVEL_ERROR,'[JustKnit] Reset Und??: $fields '. print_r($fields,1));

        $this->dictionary->fromArray($fields);
//$this->modx->log( modX::LOG_LEVEL_ERROR,'[JustKnit] Reset Und??: $errors '. print_r($this->validator->getErrors(),1));

        return $this->validator->getErrors();
    }

    /**
     * Handle the form submission, properly sanitizing and validating the data, then processing the password change
     * @return string
     */
    public function handlePost() {
        $this->dictionary = $this->register->loadDictionary();
        $this->removeSubmitVar();
//$this->modx->log( modX::LOG_LEVEL_ERROR,'[JustKnit] Reset Und??: process() $_POST '. print_r($_POST,1));

        $this->errors = $this->validate();
//$this->modx->log( modX::LOG_LEVEL_ERROR,'[JustKnit] Reset Und??: process() $this->errors '. print_r($this->errors,1));

        if (empty($this->errors)) {
//$this->modx->log( modX::LOG_LEVEL_ERROR,'[JustKnit] Reset Und??: process() $this->errors ist empty');

            if ($this->loadPreHooks()) {
//$this->modx->log( modX::LOG_LEVEL_ERROR,'[JustKnit] Reset Und??: process() $this->loadPreHooks()');
                $this->validateOldPassword();
                $this->validatePasswordLength();
                $this->confirmMatchedPasswords();
                if (empty($this->errors)) {
                    $this->changePassword();
                } else {
                    $errorMsg = $this->prepareFailureMessage();
                    $placeholderPrefix = rtrim($this->getProperty('placeholderPrefix', 'justknit.'), '.');
                    $this->modx->toPlaceholder('error_message', $errorMsg, $placeholderPrefix);
                }
            }
        }
        $this->setToPlaceholders();
        return '';
    }

    /**
     * Actually attempt to change the password
     *
     * @return boolean
     */
    public function changePassword() {
        $fieldNewPassword = $this->getProperty('fieldNewPassword');
        $fieldOldPassword = $this->getProperty('fieldOldPassword');
        $validateOldPassword = $this->getProperty('validateOldPassword',true,'isset');
        $newPassword = $this->dictionary->get($fieldNewPassword);
        $oldPassword = $this->dictionary->get($fieldOldPassword);

        /* attempt to change the password */
        $success = $this->modx->user->changePassword($newPassword,$oldPassword,$validateOldPassword);
        if (!$success) {
            /* for some reason it failed (possibly a plugin) so send error message */
            $this->errors[$fieldNewPassword] = $this->modx->lexicon('justknit.password_err_change');
        } else {
            $this->loadPostHooks();

            if (!$this->reloadOnSuccess()) {
                $this->setSuccessMessagePlaceholder();
            }
        }
        return $success;
    }

    /**
     * Set fields and errors to placeholders for a form reload
     * 
     * @return void
     */
    public function setToPlaceholders() {
        $placeholderPrefix = rtrim($this->getProperty('placeholderPrefix', 'justknit.'), '.');
        $errorPrefix = ($placeholderPrefix) ? $placeholderPrefix . '.error' : 'error';

        $fields = $this->dictionary->toArray();
        $fields = $this->escapePlaceholders($fields);

        $this->modx->toPlaceholders($this->errors, $errorPrefix);
        $this->modx->toPlaceholders($fields, $placeholderPrefix);
    }

    /**
     * Load any pre-password-change preHooks that can stop the event propagation
     * @return boolean
     */
    public function loadPreHooks() {
        $passed = true;
        $this->register->loadHooks('preHooks');
        $preHooks = $this->getProperty('preHooks','');
        if (!empty($preHooks)) {
            $this->preHooks->loadMultiple($preHooks,$this->dictionary->toArray(),array(
                'user' => &$this->modx->user,
                'submitVar' => $this->getProperty('submitVar'),
                'reloadOnSuccess' => $this->getProperty('reloadOnSuccess'),
                'fieldOldPassword' => $this->getProperty('fieldOldPassword'),
                'fieldNewPassword' => $this->getProperty('fieldNewPassword'),
                'fieldConfirmNewPassword' => $this->getProperty('fieldConfirmNewPassword'),
            ));
            $values = $this->preHooks->getValues();
            if (!empty($values)) {
                $this->dictionary->fromArray($values);
            }
        }
        /* process preHooks */
        if ($this->register->preHooks->hasErrors()) {
            $placeholderPrefix = rtrim($this->getProperty('placeholderPrefix', 'justknit.'), '.');
            $errorPrefix = ($placeholderPrefix) ? $placeholderPrefix . '.error' : 'error';
            $this->modx->toPlaceholders($this->preHooks->getErrors(), $errorPrefix);
            $this->modx->toPlaceholder('error_message', $this->preHooks->getErrorMessage(), $placeholderPrefix);
            $passed = false;
        }
        return $passed;
    }

    /**
     * If wanted, ensure the old password is validated against the current user's old password
     * @return boolean
     */
    public function validateOldPassword() {
        $validated = true;
        /* if changing the password */
        if ($this->getProperty('validateOldPassword',true,'isset')) {
            $fields = $this->dictionary->toArray();
            $fieldOldPassword = $this->getProperty('fieldOldPassword','password_old');
            
            $version = $this->modx->getVersionData();
            if (version_compare($version['full_version'],'2.1.0','>=')) {
                if (empty($fields[$fieldOldPassword]) || !$this->modx->user->passwordMatches($fields[$fieldOldPassword])) {
                    $this->errors[$fieldOldPassword] = $this->modx->lexicon('justknit.password_invalid_old');
                    $validated = false;
                }
            } else {
                if (empty($fields[$fieldOldPassword]) || md5($fields[$fieldOldPassword]) != $this->modx->user->get('password')) {
                    $this->errors[$fieldOldPassword] = $this->modx->lexicon('justknit.password_invalid_old');
                    $validated = false;
                }
            }
        }
        return $validated;
    }

    /**
     * Ensure the new password is at least the minimum length as specified in System Settings
     * 
     * @return boolean
     */
    public function validatePasswordLength() {
        $validated = true;
        $fieldNewPassword = $this->getProperty('fieldNewPassword','password_new');
        $newPassword = $this->dictionary->get($fieldNewPassword);

        $minLength = $this->modx->getOption('password_min_length',null,8);
        if (empty($newPassword) || strlen($newPassword) < $minLength) {
            $this->errors[$fieldNewPassword] = $this->modx->lexicon('justknit.password_too_short',array('length' => $minLength));
            $validated = false;
        }
        return $validated;
    }

    /**
     * If set, confirm that the confirmation password matches the new password
     * @return boolean
     */
    public function confirmMatchedPasswords() {
        $validated = true;
        $fieldConfirmNewPassword = $this->getProperty('fieldConfirmNewPassword');
        /* if using confirm, ensure they match */
        if (!empty($fieldConfirmNewPassword)) {
            $confirmNewPassword = $this->dictionary->get($fieldConfirmNewPassword);
            $fieldNewPassword = $this->getProperty('fieldNewPassword','password_new');
            $newPassword = $this->dictionary->get($fieldNewPassword);
            if (empty($confirmNewPassword) || $newPassword != $confirmNewPassword) {
                $this->errors[$fieldConfirmNewPassword] = $this->modx->lexicon('justknit.password_no_match');
                $validated = false;
            }
        }
        return $validated;
    }

    /**
     * Do any post-password-changing hooks
     * @return void
     */
    public function loadPostHooks() {
        $postHooks = $this->getProperty('postHooks','');
        if (!empty($postHooks)) {
            $placeholderPrefix = rtrim($this->getProperty('placeholderPrefix', 'justknit.'), '.');
            $errorPrefix = ($placeholderPrefix) ? $placeholderPrefix . '.error' : 'error';
            $this->loadHooks('postHooks');
            $fields['changepassword.user'] = &$this->modx->user;
            $fields['changepassword.profile'] =& $this->profile;
            $fields['changepassword.fieldOldPassword'] = $this->getProperty('fieldOldPassword');
            $fields['changepassword.fieldNewPassword'] = $this->getProperty('fieldNewPassword');
            $fields['changepassword.fieldConfirmNewPassword'] = $this->getProperty('fieldConfirmNewPassword');
            $this->postHooks->loadMultiple($postHooks,$this->dictionary->toArray());

            /* process post hooks errors */
            if ($this->postHooks->hasErrors()) {
                $this->modx->toPlaceholders($this->postHooks->getErrors(), $errorPrefix);

                $errorMsg = $this->postHooks->getErrorMessage();
                $this->modx->toPlaceholder('error_message', $errorMsg, $placeholderPrefix);
            }
        }
    }

    /**
     * If desired, reload this page on success
     * @return mixed
     */
    public function reloadOnSuccess() {
        $reloadOnSuccess = $this->getProperty('reloadOnSuccess',true,'isset');
        if ($reloadOnSuccess) {
            /* if reloading the page after success */
            $url = $this->modx->makeUrl($this->modx->resource->get('id'),'',array(
                $this->getProperty('reloadOnSuccessVar') => 1,
            ),'full');
            $this->modx->sendRedirect($url);
        }
        return $reloadOnSuccess;
    }

    /**
     * Set a success message placeholder
     * @return void
     */
    public function setSuccessMessagePlaceholder() {
        $placeholderPrefix = rtrim($this->getProperty('placeholderPrefix', 'justknit.'), '.');
        $this->modx->toPlaceholder('passwordChanged', true, $placeholderPrefix);
        $successMessage = $this->getProperty('successMessage');
        if (!empty($successMessage)) {
            $this->modx->toPlaceholder('successMessage', $successMessage, $placeholderPrefix);
        }
    }

    /**
     * @param string $defaultErrorMessage
     * @return string
     */
    public function prepareFailureMessage($defaultErrorMessage = '') {
        $errorOutput = '';
        $errTpl = $this->getProperty('errTpl');
        $errors = $this->errors;
        if (!empty($errors)) {
            foreach ($errors as $error) {
                $errorOutput .= $this->modx->getChunk($errTpl,  array('msg' => $error));
            }
        } else {
            $errorOutput = $this->modx->getChunk($errTpl, array('msg' => $defaultErrorMessage));
        }
        return $errorOutput;
    }    
}
